
from helper import *
from math import log
from random import randint

# Modular inverse of an integer
# Source: https://en.wikibooks.org/wiki/Algorithm_Implementation/Mathematics/Extended_Euclidean_algorithm
def egcd(a, b):
    if a == 0:
        return (b, 0, 1)
    else:
        g, y, x = egcd(b % a, a)
        return (g, x - (b // a) * y, y)

def modinv(a, m):
    g, x, y = egcd(a, m)
    if g != 1:
        raise Exception('Modular inverse does not exist.')
    else:
        return x % m

# Bit-Reverse an integer
def int_bitreverse(a,l):
    b = ('{:0'+str(l)+'b}').format(a)
    return int(b[::-1],2)

# Bit-Reverse an array
def arr_reverse(a):
    l = int(log(len(a),2)) 
    b = [0]*len(a)
    for i in range(len(a)):
        i_rev = int_bitreverse(i,l)
        b[i_rev] = a[i]
    return b

# Check if input w is m-th (could be n or 2n) primitive root of unity of q
def is_primitive_root(w,m,q):
    if w == 0:
        return False
    elif pow(w,m//2,q) == (q-1):
        return True
    else:
        return False

# Returns a m-th primitive root of q (if there is any)
def find_primitive_root(m,q):
    g = (q-1)//m

    if (q-1) != g*m:
        return False,0

    attempt_ctr = 0
    attempt_max = 100

    while(attempt_ctr < attempt_max):
        a = randint(2,q-1)
        b = pow(a,g,q)
        # check
        if is_primitive_root(b,m,q):
            return True,b
        else:
            attempt_ctr = attempt_ctr+1

    return False,0

# Returns the minimum m-th primitive root of q (if there is any)
def find_min_primitive_root(m,q):
    p_found, g = find_primitive_root(m,q)

    if not(p_found):
        return False,0

    g_sq = (g*g) % q
    cg   = g

    for i in range(m):
        if (cg < g):
            g = cg
        cg = (cg*g_sq) % q

    return True,g

# Generate necessary BFV parameters given n and log(q)
def generate_ntt_parameters(q,n):
    if q % (2*n) != 1:
        raise Exception("Given coefficient modulus does not meet requirement for NTT.")
    else:
        q_found, psi = find_primitive_root(2*n,q)
        if not(q_found):
            raise Exception("Cannot find a primitive root for given coefficient modulus.")
        else:
            psi_inv= modinv(psi,q)
    n_inv = modinv(n,q)

    # create tables
    psi_table = [1]*n
    psi_inv_table = [1]*n

    for i in range(1,n):
        psi_table[i] = (psi_table[i-1]*psi) % q
        psi_inv_table[i] = (psi_inv_table[i-1]*psi_inv) % q

    return psi_table, psi_inv_table, n_inv
