
from math import log
from helper import *

# Forward NTT
# Input : In normal order
# Output: In bit-reversed order
def fntt(A,psi_table,q):
    N = len(A)
    B = [_ for _ in A] # create a copy of input since this operation is in-place

    l = int(log(N,2))

    t = N
    m = 1
    while(m<N):
        t = int(t/2)
        for i in range(m):
            j1 = 2*i*t
            j2 = j1 + t - 1
            psi_pow = int_bitreverse(m+i,l)
            
            W = psi_table[psi_pow]

            for j in range(j1,j2+1):
                # Cooley-Tukey butterfly operation
                U = B[j]
                V = (B[j+t]*W) % q

                B[j]   = (U+V) % q
                B[j+t] = (U-V) % q
        m = 2*m

    return B

# Inverse NTT
# Input : In bit-reversed order
# Output: In normal order
def intt(A,psi_inv_table,q,n_inv):
    N = len(A)
    B = [_ for _ in A] # create a copy of input since this operation is in-place

    l = int(log(N,2))

    t = 1
    m = N
    while(m>1):
        j1 = 0
        h = int(m/2)
        for i in range(h):
            j2 = j1 + t - 1
            psi_pow = int_bitreverse(h+i,l)
            
            W = psi_inv_table[psi_pow]
            
            for j in range(j1,j2+1):
                # Gentleman-Sande butterfly operation
                U = B[j]
                V = B[j+t]

                B[j]   = (U+V) % q
                B[j+t] = (U-V)*W % q
            j1 = j1 + 2*t
        t = 2*t
        m = int(m/2)

    # Final scaling (multiplication with 1/n)
    for i in range(N):
        B[i] = (B[i] * n_inv) % q

    return B

# Schoolbook-based polynomial multiplicatin in ring Z_q[x]/x^n+1
def schoolbook_polmul(A, B, q):
    C = [0] * (2 * len(A))
    D = [0] * (len(A))
    for indexA, elemA in enumerate(A):
        for indexB, elemB in enumerate(B):
            C[indexA + indexB] = (C[indexA + indexB] + elemA * elemB) % q

    for i in range(len(A)):
        D[i] = (C[i] - C[i + len(A)]) % q
    return D