from helper import *
from ntt import *
from random import randint
from math import log

# ------------------------ Set your parameters
n = 256
q = 18434813901432784897

# ------------------------ Generate NTT parameters
psi_table, psi_inv_table, n_inv = generate_ntt_parameters(q,n)

# ------------------------ Testing FNTT-INTT conversion
A = [randint(0,q-1) for _ in range(n)]

# FNTT-INTT operations
A_ntt = fntt(A,psi_table,q)
A_pol = intt(A_ntt,psi_inv_table,q,n_inv)

# Comparison
if sum([abs(p0-p1) for p0,p1 in zip(A,A_pol)]) == 0:
    print("FNTT-INTT conversion worked!")
else:
    print("FNTT-INTT conversion failed!")

# ------------------------ Testing NTT-based polynomial multiplication
A = [randint(0,q-1) for _ in range(n)]
B = [randint(0,q-1) for _ in range(n)]

# Schoolbook-based multiplication
C = schoolbook_polmul(A,B,q)

# NTT-based multiplication
A_ntt = fntt(A,psi_table,q)
B_ntt = fntt(B,psi_table,q)

C_ntt = [(a*b)%q for a,b in zip(A_ntt,B_ntt)]

C_pol = intt(C_ntt,psi_inv_table,q,n_inv)

# Comparison
if sum([abs(p0-p1) for p0,p1 in zip(C,C_pol)]) == 0:
    print("Polynomial multiplication worked!")
else:
    print("Polynomial multiplication failed!")

